package webwisdom.tango.object;

import java.util.*;

/**
  * An internal alarm entry class for storing information about alarms.
  */
class AlarmEntry
{
  AlarmEntry(long time,AlarmListener a,AlarmEvent av)
    {
      this.time=time;
      this.a=a;
      this.av=av;
    }
  long 		time;
  AlarmListener a;
  AlarmEvent    av;
}

/**
  * A general purpose alarm class.
  * 
  * @author Piotr Sokolowski
  * @version $Revision: 1.3 $
  * @see webwisdom.tango.object.AlarmListener
  * @see webwisdom.tango.object.AlarmEvent
  */
public class Alarm implements Runnable
{
  /**
    * the single instance of the Alarm class
    */
  public static Alarm alarmInstance=new Alarm();

  private Thread th;
  private boolean running;  private Vector v;
  private AlarmEntry next;
  /**
    * a private constructor. This ensures that there would be only one instance of Alarm running.
    */
  private Alarm()
    {
      v=new Vector();
      next=null;
      running=true;
      th=new Thread(this);
      th.start();
    }

  /**
    * adds a new entry to the alarms list
    * @param time the time (as in system time) at which the alarm should go off.
    * @param al listener which should be notified about the alarm event
    * @param av event which will be passed to the AlarmListener when the time comes. :)
    */
  synchronized public void add(long time, AlarmListener al,AlarmEvent av)
    {
      AlarmEntry ae=new AlarmEntry(time,al,av);
      v.addElement(ae);
      notify();
    }

  /**
    * removes all entries associated with a given AlarmListener from the alarm queue.
    * @param al listener
    */
  synchronized public void remove(AlarmListener al)
    {
      Vector nv=new Vector();
      for(int i=0;i<v.size();i++)
	if(!((AlarmEntry)v.elementAt(i)).a.equals(al))
	  nv.addElement(v.elementAt(i));
      v=nv;
      if(next!=null && next.a.equals(al)) next=null;
      notify();
    }

  /**
    * removes all instances where given AlarmListener is notified by the give AlarmEvent
    * @param a AlarmListener
    * @param av AlarmEvent
    */
  synchronized public void remove(AlarmListener a, AlarmEvent av)
    {
      Vector nv=new Vector();
      for(int i=0;i<v.size();i++)
	{
	  AlarmEntry ae=(AlarmEntry)v.elementAt(i);
	  if(!ae.a.equals(a) || !ae.av.equals(av))
	    nv.addElement(v.elementAt(i));
	}
      v=nv;
      if(next!=null && next.a.equals(a) && next.av.equals(av)) next=null;
      notify();
    }

  /**
    * gets current time
    * @return current time
    */
  synchronized public long getTime()
    {
      return System.currentTimeMillis();
    }

  /**
    * does not belong to the API
    */
  synchronized public void run()
    {
      while(running)
	{
	  try
	    {
	      next=null;
	      for(int i=0;i<v.size();i++)
		{
		  AlarmEntry ae=(AlarmEntry)v.elementAt(i);
		  if(next==null || ae.time<next.time) next=ae;
		}
	      if(next==null)
		wait();
	      else
		{
		  long t=next.time-getTime();
		  if(t>0) wait(t);
		}
	      if(next!=null && next.time<getTime())
		{
		  v.removeElement(next);
		  next.a.processAlarm(next.av);
		}
	    }
	  catch(Exception e)
	    {
	      e.printStackTrace();
	    }
	}
    }

  /**
    * destroys the alarm object
    * stops the thread etc...
    */
  synchronized public void dispose()
    {
      running=false;
      next=null;
      notify();
    }
}
